unit Fingers;

interface

	uses
		MyOOMainLoop;

	const
		WT_FingerOutput = 'FOut';

	type
		DefObject = object(DObject)
				procedure CalculateRegion (var rgn: rgnHandle);
				override;
				function DoMainClick (er: eventRecord; wp: windowPtr; code: integer): boolean;
				override;
			end;

	var
		giveOoMerror: boolean;   { Enable out of mem error every time a menu is chosen }

	procedure InitFingers;
	procedure FinishFingers;
	procedure HandleFingerTCPEvents;
	procedure Finger;
	procedure ParseURL (var s: str255; whois: boolean);
	procedure DoFingerCommand (s: str255);
	procedure FailAlert (index: integer; s1: str255; n: longInt);
	function IsFingerWindow (wp: windowPtr): boolean;
	function GetFingerTE (wp: windowPtr): TEHandle;

implementation

	uses
		ToolUtils, MySystemGlobals, AppGlobals, Preferences, TranslateISO, MyAEUtils, AppleEvents, QAlerts, MyStandardFile,{}
		TCPTypes, TCPStuff, TCPConnections, OOStaticEdit, BaseGlobals, FingerRegistry, MyStrings, MyFileSystemUtils,{}
		MyInternetMenu, MyAddressInput, MyGrowZones, MyDialogs, MyTypes, PrefCodes, MyNewPreferences;

	const
		my_waiting_cursor = 128;

	type
		FIObject = object(AIObject)
				procedure Create (id: integer);
				override;
				procedure DoItem (item: integer);
				override;
				procedure DoKey (modifiers: integer; ch: char; code: integer);
				override;
				procedure PackStringBlank (var s: str255);
				override;
				procedure PackString (var s: str255);
				override;
				procedure UnPackString (s: str255);
				override;
				function IsOKOn: boolean;
				override;
				function IsSetDefaultOn: boolean;
				override;
				procedure DoCommand (s: str255);
				override;
				procedure GetDialog;
				override;
			end;
		FOObject = object(DObject)
				textob: TEStaticObject;
				procedure Create (id: integer);
				override;
				procedure Destroy;
				override;
				procedure Resize;
				override;
				function EditMenuEnabled: boolean;
				override;
				procedure SetEditMenuItem (item: integer);
				override;
				procedure DoEditMenu (item: integer);
				override;
				procedure Zoom (zoomout: boolean; fullscreen: boolean);
				override;
				function DoMainClick (er: eventRecord; wp: windowPtr; code: integer): boolean;
				override;
				procedure DoItemWhere (er: eventRecord; item: integer);
				override;
				procedure DoActivateDeactivate (activate: boolean);
				override;
				procedure CalculateRegion (var rgn: rgnHandle);
				override;
			end;

	type
		FingerDestination = (FD_None, FD_Window, FD_File, FD_Reply);

	type
		FingerInfoRecord = record
				what: str255;
				se: SuspendedEvent;
				dest: FSSpec;
				fd: FingerDestination;
				foo: FOObject;
				datahandle: handle;
				remoteip: longInt;
			end;
		FingerInfoPtr = ^FingerInfoRecord;

	const
		finger_input_refcon = -1;
		fi_port = 7;
		fi_save = 8;
		fo_text_item = 1;
		ports_strh_id = 300;
		finger_str_index = 1;
		finger_port_index = 2;
		whois_str_index = 3;
		whois_port_index = 4;
		general_strh_id = 400;
		cmd_key_index = 2;
		finger_button_index = 3;
		whois_button_index = 4;
		max_sane_handle_size = 32000;

	var
		default_whois: boolean;
		fingering: integer;
		finger_port, whois_port: integer;
		whois_str: str63;
		cmd_key: char;
		finger_button_str, whois_button_str: string[10];

	procedure FailAlert (index: integer; s1: str255; n: longInt);
		var
			s, s2: str255;
	begin
		if n = 0 then begin
			s2 := '';
		end
		else begin
			NumToString(n, s2);
		end;
		GetIndString(s, 150, index);
		SPrintS3(s, s, s1, s2, '');
		QParamText(s, '', '', '');
		QStopAlert(fail_alert_id, nil);
	end;

	procedure PackName (var s: str255; name, mach: str255; whois: boolean);
		function hasat (n: str255): boolean;
			var
				i: integer;
		begin
			hasat := Pos('@', n) > 0;
		end;

		procedure Validate (n: str255);
			var
				i: integer;
		begin
			s := n;
			if s[length(s)] = '@' then
				s := '?';
			i := 1;
			while i < length(s) do begin
				if (s[i] = '@') and (s[i + 1] = '@') then
					s := '?';
				i := i + 1;
			end;
			if Pos(concat(':', whois_str), s) <> 0 then
				s := '?';
		end;
	begin
		s := '?';
		if mach = '' then begin
			mach := name;
			name := '';
		end;
		if (name = '') then begin
			mach := Trim(mach);
			if (length(mach) > 1) & (mach[1] = '<') & (mach[length(mach)] = '>') then begin
				Delete(mach, 1, 1);
				Delete(mach, length(mach), 1);
			end;
			if IUEqualString(copy(mach, 1, 4), 'url:') = 0 then begin
				Delete(mach, 1, 4);
			end;
			if (IUEqualString(copy(mach, 1, 7), 'finger:') = 0) | (IUEqualString(copy(mach, 1, 6), 'whois:') = 0) then begin
				whois := IUEqualString(mach[1], 'w') = 0;
				Delete(mach, 1, 6 + ord(not whois));
				if copy(mach, 1, 2) = '//' then begin
					Delete(mach, 1, 2);
				end;
				if not hasat(mach) then begin
					SplitBy(mach, '/', mach, name);
				end;
			end;
		end;
		if hasat(name) then begin
			if mach = '' then
				Validate(name);
		end
		else if hasat(mach) then begin
			if name = '' then
				Validate(mach);
		end
		else begin
			Validate(concat(name, '@', mach));
		end;
		if (s <> '?') and whois then
			s := concat(s, ':', whois_str);
	end;

	procedure UnpackName (s: str255; var name, mach: str255; var whois: boolean);
		var
			p: integer;
	begin
		p := length(s);
		while s[p] <> '@' do
			p := p - 1;
		name := copy(s, 1, p - 1);
		mach := copy(s, p + 1, 255);
		p := Pos(concat(':', whois_str), mach);
		whois := p > 0;
		if whois then
			mach := copy(mach, 1, p - 1);
	end;

	procedure ParseURL (var s: str255; whois: boolean);
	begin
		PackName(s, '', s, whois);
	end;

	function IsFingerWindow (wp: windowPtr): boolean;
	begin
		IsFingerWindow := GetWType(wp) = WT_FingerOutput;
	end;

	function GetFingerTE (wp: windowPtr): TEHandle;
	begin
		GetFingerTE := FOObject(GetWObject(wp)).textob.te;
	end;

	procedure FIObject.PackStringBlank (var s: str255);
	begin
		if (default_machine = '') and (default_user = '') then
			if default_whois then
				s := concat('@:', whois_str) {sleeze}
			else
				s := '@'
		else
			PackName(s, default_user, default_machine, default_whois);
	end;

	procedure FIObject.PackString (var s: str255);
	begin
		PackName(s, default_user, default_machine, default_whois);
	end;

	procedure FIObject.UnPackString (s: str255);
	begin
		UnPackName(s, default_user, default_machine, default_whois);
	end;

	function FIObject.IsOKOn: boolean;
		var
			s, s1, s2: str255;
	begin
		GetItemText(window, ai_user, s1);
		GetItemText(window, ai_machine, s2);
		PackName(s, s1, s2, false);
		IsOkOn := s <> '?';
	end;

	function FIObject.IsSetDefaultOn: boolean;
		var
			s, s1, s2: str255;
	begin
		GetItemText(window, ai_user, s1);
		GetItemText(window, ai_machine, s2);
		PackName(s, s1, s2, false);
		IsSetDefaultOn := (s <> '?') or ((s1 = '') and (s2 = ''));
	end;

	procedure FIObject.GetDialog;
		var
			kind: integer;
			h: handle;
			r: rect;
	begin
		inherited GetDialog;
		GetDialogItem(window, fi_port, kind, h, r);
		default_whois := GetControlValue(controlhandle(h)) <> 0;
	end;

	procedure TogglePort (wp: windowPtr);
		var
			kind: integer;
			h: controlHandle;
			r: rect;
			cv: integer;
	begin
		GetDialogItem(wp, fi_port, kind, handle(h), r);
		cv := 1 - GetControlValue(h);
		SetControlValue(h, cv);
		GetDialogItem(wp, ai_ok, kind, handle(h), r);
		if cv = 0 then {finger}
			SetControlTitle(h, finger_button_str)
		else
			SetControlTitle(h, whois_button_str);
	end;

	procedure FIObject.Create (id: integer);
		var
			kind: integer;
			h: handle;
			r: rect;
	begin
		inherited Create(id);
		GetDialogItem(window, fi_port, kind, h, r);
		SetControlValue(controlhandle(h), ord(default_whois));
		TogglePort(window); { Set up Finger/Whois button }
		TogglePort(window);
		ShowWindow(window);
	end;

	procedure CreateFingerInfo (var p: FingerInfoPtr);
	begin
		p := FingerInfoPtr(NewPtr(SizeOf(FingerInfoRecord)));
		p^.fd := FD_None;
		p^.se := NullSuspendedEvent;
		p^.datahandle := NewHandle(0);
		fingering := fingering + 1;
	end;

	procedure DestroyFingerInfo (var p: FingerInfoPtr);
	begin
		if p <> nil then begin
			ResumeEvent(p^.se);
			DisposeHandle(p^.datahandle);
			DisposePtr(Ptr(p));
			fingering := fingering - 1;
			p := nil;
		end;
	end;

	procedure ResumeFingerInfoError (p: FingerInfoPtr; err: OSErr);
		var
			junk: OSErr;
	begin
		if (p <> nil) & (p^.se.waiting) then begin
			if (p^.fd = FD_Reply) then begin
				if (err <> noErr) then begin
					junk := AEPutParamPtr(p^.se.reply, keyErrorNumber, typeShortInteger, @err, SizeOf(err));
				end;
			end
			else begin
				junk := AEPutParamPtr(p^.se.reply, keyDirectObject, typeShortInteger, @err, SizeOf(err));
			end;
		end;
		DestroyFingerInfo(p);
	end;

	procedure ResumeFingerInfoReply (p: FingerInfoPtr);
		var
			junk, err: OSErr;
			retval: AEDesc;
	begin
		err := -1;
		if (p <> nil) & (p^.se.waiting) then begin
			HLock(p^.datahandle);
			err := AEPutParamPtr(p^.se.reply, keyDirectObject, typeChar, p^.datahandle^, GetHandleSize(p^.datahandle));
		end;
		ResumeFingerInfoError(p, err);
	end;

	procedure DoRealFingerCommand (p: fingerInfoPtr);
		var
			cp: connectionIndex;
			oe: OSErr;
			s1, s2: str255;
			whois: boolean;
	begin
		UnpackName(p^.what, s1, s2, whois);
		oe := FindAddress(cp, s2, p);
		if oe <> noErr then begin
			ResumeFingerInfoError(p, oe);
			FailAlert(1, '', oe);
		end;
	end;

	const
		R_Finger = 0;
		R_Whois = 1;
		R_GURL = 2;
		R_FURL = 3;

	function HandleAEFinger (event, reply: AppleEvent; refcon: longInt): OSErr;
		var
			err, oerr, junk: OSErr;
			s: str255;
			p: FingerInfoPtr;
	begin
		CreateFingerInfo(p);
		p^.fd := FD_Window;
		err := noErr;
		if not EnoughSpace(0, 12000) then begin
			err := memFullErr;
		end;
{required}
		if err = noErr then begin
			err := GetStringFromAERecord(event, keyDirectObject, s);
		end;
{optional}
		if (err = noErr) & (refcon = R_GURL) then begin
			oerr := GetFSSpecFromAERecord(event, keyAEDestination, p^.dest);
			if oerr = noErr then begin
				p^.fd := FD_File;
			end;
		end;
		if err = noErr then
			err := GotRequiredParams(event);
		if err = noErr then
			err := SuspendEvent(event, reply, nil, 0, p^.se);
		if err = noErr then begin
			case refcon of
				R_Finger:  begin
					ParseURL(s, false);
				end;
				R_Whois:  begin
					ParseURL(s, true);
				end;
				R_GURL:  begin
					ParseURL(s, false);
				end;
				R_FURL:  begin
					ParseURL(s, false);
					p^.fd := FD_Reply;
				end;
			end;
			p^.what := s;
			DoRealFingerCommand(p);
		end
		else begin
			DestroyFingerInfo(p);
		end;
		HandleAEFinger := err;
	end;

	procedure DoFingerCommand (s: str255);
		var
			event, reply: AppleEvent;
			target: AEDesc;
			s1, s2: str255;
			whois: boolean;
			err: OSErr;
			p: FingerInfoPtr;
	begin
		if has_AppleEvents then begin
			UnpackName(s, s1, s2, whois);
			AECreate(reply);
			err := CreateSelfTarget(target);
			if whois then begin
				err := AECreateAppleEvent(kMySuite, kAEWhois, target, kAutoGenerateReturnID, kAnyTransactionID, event);
			end
			else begin
				err := AECreateAppleEvent(kMySuite, kAEFinger, target, kAutoGenerateReturnID, kAnyTransactionID, event);
			end;
			AEDestroy(target);
			err := PutStringToAERecord(event, keyDirectObject, concat(s1, '@', s2));
			err := AESend(event, reply, kAENoReply + kAEAlwaysInteract, kAENormalPriority, kAEDefaultTimeOut, nil, nil);
		end
		else begin
			CreateFingerInfo(p);
			p^.fd := FD_Window;
			p^.what := s;
			DoRealFingerCommand(p);
		end;
	end;

	procedure FIObject.DoCommand (s: str255);
	begin
		DoFingerCommand(s);
	end;

	procedure FIObject.DoItem (item: integer);
		var
			reply: MySFreply;
			rn: integer;
			err, oerr: OSErr;
			s: str255;
	begin
		case item of
			fi_port: 
				TogglePort(window);
			fi_save:  begin
				PutFile('', '', reply);
				if reply.Rgood then begin
					err := HDelete(reply.RvrefNum, reply.RdirID, reply.Rfname);
					err := HCreate(reply.RvrefNum, reply.RdirID, reply.Rfname, 'PnLF', 'AURL');
					err := HOpen(reply.RvrefNum, reply.RdirID, reply.Rfname, fsWrPerm, rn);
					if err = noErr then begin
						GetDialog;
						PackString(s);
						s := concat('FCH ', s);
						err := MyFSWrite(rn, length(s), @s[1]);
						oerr := FSClose(rn);
						if err = noErr then begin
							err := oerr;
						end;
					end;
					FailAlert(2, '', err);
				end;
			end;
			otherwise
				inherited DoItem(item);
		end;
	end;

	procedure FIObject.DoKey (modifiers: integer; ch: char; code: integer);
	begin
		if (BAND(modifiers, cmdKey) <> 0) and (IUEqualString(ch, cmd_key) = 0) then
			TogglePort(window);
		inherited DoKey(modifiers, ch, code);
	end;

	procedure FailOutOfMemory;
	begin
		if giveOoMerror then begin
			FailAlert(3, '', 0);
			giveOoMerror := false;
		end;
	end;

	function ReadChars (tcpc: TCPConnectionPtr; h: handle; count: longInt): boolean;
{ return true if out of memory }
		var
			size, i, j: longInt;
			b: signedByte;
			p: ptr;
			oe: OSErr;
	begin
		ReadChars := false;
		if MemoryCritical then begin
			FailOutOfMemory;
			oe := TCPFlush(tcpc);
			ReadChars := true;
			Exit(ReadChars);
		end;
		size := GetHandleSize(h);
		if size + count > max_sane_handle_size then
			count := max_sane_handle_size - size;
		if count > 0 then begin
			SetHandleSize(h, size + count);
			if GetHandleSize(h) <> size + count then begin
				FailOutOfMemory;
				oe := TCPFlush(tcpc);
				ReadChars := true;
				Exit(ReadChars);
			end
			else begin
				HLock(h);
				oe := TCPReceiveChars(tcpc, ptr(longInt(h^) + size), count);
				if oe = noErr then begin
					count := NetToMac(ptr(longInt(h^) + size), count);
				end;
				HUnlock(h);
				SetHandleSize(h, size + count);
			end;
		end
		else
			oe := TCPFlush(tcpc);
	end;

	function StripCRLF (h: handle): boolean;
		var
			size, i: longInt;
			b: signedByte;
	begin
		StripCRLF := false;
		size := GetHandleSize(h);
		for i := 0 to size - 1 do begin
			b := ptr(longInt(h^) + i)^;
			if (b = 13) or (b = 10) then begin
				StripCRLF := true;
				SetHandleSize(h, i);
				leave;
			end;
		end;
	end;

{$S Init}
	procedure InitFingers;
		procedure GetIndNumber (id, index: integer; var num: integer);
			var
				temp: longInt;
				s: str255;
		begin
			GetIndString(s, id, index);
			StringToNum(s, temp);
{$PUSH}
{$R-}
			num := temp;
{$R-}
		end;
		var
			th: handle;
			i: integer;
			s: str255;
			junk: OSErr;
	begin
		fingering := 0;
		GetIndNumber(ports_strh_id, finger_port_index, finger_port);
		GetIndNumber(ports_strh_id, whois_port_index, whois_port);
		GetIndString(s, ports_strh_id, whois_str_index);
		whois_str := s;
		GetIndString(s, general_strh_id, cmd_key_index);
		cmd_key := s[1];
		GetIndString(s, general_strh_id, finger_button_index);
		finger_button_str := s;
		GetIndString(s, general_strh_id, whois_button_index);
		whois_button_str := s;
		InitAddressInput;
		if has_AppleEvents then begin
			junk := AEInstallEventHandler(kMySuite, kAEFinger, @HandleAEFinger, R_Finger, false);
			junk := AEInstallEventHandler(kMySuite, kAEWhois, @HandleAEFinger, R_Whois, false);
			junk := AEInstallEventHandler('GURL', 'GURL', @HandleAEFinger, R_GURL, false);
			junk := AEInstallEventHandler('GURL', 'FURL', @HandleAEFinger, R_FURL, false);
		end;
	end;

{$S Term}
	procedure FinishFingers;
	begin
	end;

{$S}
	procedure FOObject.Resize;
		var
			kind, fsize, bt, rt: integer;
			h: handle;
			r: rect;
			finfo: FontInfo;
	begin
		SetPort(window);
		GetDialogItem(window, fo_text_item, kind, h, r);
		r := windowPeek(window)^.port.portRect;
		InsetRect(r, -1, -1);
		SetDialogItem(window, fo_text_item, kind, h, r);
		textob.Resize;
	end;

	procedure FOObject.Zoom (zoomout: boolean; fullscreen: boolean);
		var
			lines: integer;
	begin
		if zoomout then begin
			with textob.te^^ do begin
				lines := nLines;
		{since nLines isnt right if the last character is a return, check for that case}
				if Ptr(ORD(hText^) + teLength - 1)^ = 13 then
					lines := lines + 1;
				zoomSize.v := lines * lineHeight + 20;
			end;
			if zoomSize.v < growRect.top then
				zoomSize.v := growRect.top;
			if zoomSize.v > growRect.bottom then
				zoomSize.v := growRect.bottom;
		end;
		inherited Zoom(zoomout, fullscreen);
	end;

	procedure FOObject.DoActivateDeactivate (activate: boolean);
	begin
		textob.DoActivateDeactivate(activate);
	end;

	procedure Finger;
		var
			fio: FIObject;
	begin
		new(fio);
		fio.Create(finger_input_dialog_id);
		ShowWindow(fio.window);
	end;

	function FOObject.EditMenuEnabled: boolean;
	begin
		EditMenuEnabled := textob.EditMenuEnabled;
	end;

	procedure FOObject.SetEditMenuItem (item: integer);
	begin
		textob.SetEditMenuItem(item);
	end;

	procedure FOObject.DoEditMenu (item: integer);
	begin
		textob.DoEditMenu(item);
	end;

	function FOObject.DoMainClick (er: eventRecord; wp: windowPtr; code: integer): boolean;
		var
			pt: point;
	begin
		pt := er.where;
		SetPort(window);
		GlobalToLocal(pt);
		if not PtInRect(pt, textob.te^^.viewRect) then begin
			InitCursor;
		end;
		DoMainClick := inherited DoMainClick(er, wp, code);
	end;

	procedure FOObject.DoItemWhere (er: eventRecord; item: integer);
	begin
		textob.DoItemWhere(er, item);
		if textob.te^^.selStart = textob.te^^.selEnd then begin  { kludge to make the carret go away }
			TEDeactivate(textob.te);
			TEActivate(textob.te);
		end;
	end;

	function DefObject.DoMainClick (er: eventRecord; wp: windowPtr; code: integer): boolean;
	begin
		InitCursor;
		DoMainClick := inherited DoMainClick(er, wp, code);
	end;

	procedure DefObject.CalculateRegion (var rgn: rgnHandle);
	begin
		if fingering > 0 then
			SetCursor(GetCursor(my_waiting_cursor)^^)
		else
			InitCursor;
		rgn := nil;
	end;

	procedure FOObject.CalculateRegion (var rgn: rgnHandle);
		var
			pt: point;
			rgn2: rgnHandle;
			r: rect;
	begin
		rgn := NewRgn;

		r := textob.te^^.viewRect;
		SetPort(window);
		GetMouse(pt);
		RectRgn(rgn, r);
		if PtInRect(pt, r) then begin
			SetCursor(GetCursor(iBeamCursor)^^);
		end
		else begin
			if fingering > 0 then
				SetCursor(GetCursor(my_waiting_cursor)^^)
			else
				InitCursor;
			rgn2 := NewRgn;
			SetRectRgn(rgn2, -30000, -30000, 30000, 30000);
			DiffRgn(rgn2, rgn, rgn);
			DisposeRgn(rgn2);
		end;
	end;

	procedure DrawFingerText (dp: dialogPtr; item: integer);
	begin
		FOObject(GetWObject(dp)).textob.Draw;
	end;

	procedure InsertText (sto: TEStaticObject; h: handle);
		var
			junk: OSErr;
	begin
		junk := HandAndHand(h, sto.te^^.hText);
		TECalText(sto.te);
		sto.Adjust;
	end;

	procedure FOObject.Create (id: integer);
		var
			kind, lw: integer;
			h: handle;
			r: rect;
			temptextob: TEStaticObject;
	begin
		inherited Create(id);
		window_type := WT_FingerOutput;
		SetPort(window);
		TextFont(monaco);
		TextSize(9);
		new(temptextob);
		textob := temptextob;
		lw := CharWidth('a') * 80 + 2;
		textob.Create(window, fo_text_item, lw, true, true, true, true);
		zoomSize.h := lw + 20;
		GetDialogItem(window, fo_text_item, kind, h, r);
		SetDialogItem(window, fo_text_item, kind, handle(@DrawFingerText), r);
		Resize;
	end;

	procedure FOObject.Destroy;
		var
			p: FingerInfoPtr;
	begin
		textob.Destroy;
		inherited Destroy;
	end;

	function FindPort (whois: boolean): integer;
	begin
		if whois then
			FindPort := whois_port
		else
			FindPort := finger_port;
	end;

	function AddIP (s: str255; ip: longInt): str255;
		var
			s1, s2, sip: str255;
			whois: boolean;
	begin
		if prefs.GetTagBoolean(P_ShowIP) then begin
			UnpackName(s, s1, s2, whois);
			FindString(ip, sip);
			if sip <> s2 then begin
				s2 := concat(s2, ' (', sip, ')');
				PackName(s, s1, s2, whois);
			end;
		end;
		AddIP := s;
	end;

	procedure HandleFingerTCPEvents;
		var
			oe, err, oerr: OSErr;
			cer: connectionEventRecord;
			s, s1, s2: str255;
			dlg: dialogPtr;
			cp: connectionIndex;
			texth: handle;
			remoteIP: longInt;
			defrefnum, i, ps: integer;
			whois: boolean;
			prefs_fs: FSSpec;
			prefs_rn: integer;
			n: longInt;
			p: FingerInfoPtr;
			rn: integer;
	begin
		while GetConnectionEvent(any_connection, cer) do
			with cer do begin
				p := FingerInfoPtr(dataptr);
				case event of
					C_Found:  begin
						UnpackName(p^.what, s1, s2, whois);
						p^.remoteip := value;
						oe := NewActiveConnection(cp, 0, 0, value, FindPort(whois), p);
						if oe <> noErr then begin
							FailAlert(4, s2, oe);
							ResumeFingerInfoError(p, oe);
						end;
					end;
					C_SearchFailed:  begin
						UnpackName(p^.what, s1, s2, whois);
						FailAlert(5, s2, value);
						ResumeFingerInfoError(p, value);
					end;
					C_FailedToOpen:  begin
						UnpackName(p^.what, s1, s2, whois);
						if timedout then begin
							FailAlert(6, s2, 0);
						end
						else begin
							FailAlert(7, s2, 0);
						end;
						ResumeFingerInfoError(p, -1);
						SetDataPtr(connection, nil);
					end;
					C_Established:  begin
						if MemoryCritical then begin
							FailOutOfMemory;
							ResumeFingerInfoError(p, memFullErr);
							SetDataPtr(connection, nil);
							CloseConnection(connection);
						end
						else begin
							prefs.GetTagLong(P_Fingers, n);
							prefs.SetTagLong(P_Fingers, n + 1);
							UnpackName(p^.what, s1, s2, whois);
							s1 := concat(s1, chr(13), chr(10));
							oe := TCPSend(tcpc, @s1[1], length(s1), true);
							AddInternetCommand(p^.what);
						end;
					end;
					C_CharsAvailable:  begin
						if p <> nil then begin
							if MemoryCritical then begin
								FailOutOfMemory;
								ResumeFingerInfoError(p, memFullErr);
								SetDataPtr(connection, nil);
							end
							else if ReadChars(tcpc, p^.datahandle, value) then begin
								CloseConnection(connection);
							end;
						end;
					end;
					C_Closing:  begin
						CloseConnection(connection);
					end;
					C_Closed:  begin
						if p <> nil then begin
							prefs.GetTagLong(P_Transferred, n);
							prefs.SetTagLong(P_Transferred, n + (GetHandleSize(p^.datahandle) + 1023) div 1024);
							case p^.fd of
								FD_None:  begin
									DestroyFingerInfo(p);
								end;
								FD_Window:  begin
									new(p^.foo);
									p^.foo.Create(finger_output_dialog_id);
									SetWTitle(p^.foo.window, AddIP(p^.what, p^.remoteip));
									InsertText(p^.foo.textob, p^.datahandle);
									p^.foo.InitialZoom(0, 0);
									ShowWindow(p^.foo.window);
									ResumeFingerInfoError(p, noErr);
								end;
								FD_File:  begin { You can't get here from System 6 - it requires AppleEvents }
									err := FSpDelete(p^.dest);
									err := FSpCreate(p^.dest, 'ttxt', 'TEXT', 0);
									err := FSpOpenDF(p^.dest, fsWrPerm, rn);
									if err = noErr then begin
										HLock(p^.datahandle);
										err := MyFSWrite(rn, GetHandleSize(p^.datahandle), p^.datahandle^);
										oerr := FSClose(rn);
										if err = noErr then begin
											err := oerr;
										end;
									end;
									ResumeFingerInfoError(p, noErr);
								end;
								FD_Reply:  begin
									ResumeFingerInfoReply(p);
								end;
							end;
							if p^.foo <> nil then begin
							end
							else begin
								ResumeFingerInfoError(p, -1);
							end;
						end;
					end;
					otherwise
						;
				end;{case}
			end;{while with}
	end;

end.