unit Preferences;

interface

	procedure InitPreferencesR;
	function FinishPreferencesR: boolean;

implementation

	uses
		MyTypes, MyUtils, MySystemGlobals, MyStandardFile, Folders, MyMenus, MyVersionResource, {}
		PrefCodes, MyCollections,{}
		BaseGlobals, AppGlobals, MyOOMainLoop, MyOOAboutHelp,{}
		MyFMenus, MyNewPreferences, MyDialogs, MyFileSystemUtils;

	const
		Pref_Save_item = 1;
		Pref_Default_Item = Pref_Save_item;  { Must be one anyway, since we use OutlineDefault1 }
		Pref_Cancel_item = 2;

		User_item = 3;
		IPS_item = 4;
		AFS_item = 5;
		Paid_item = 6;

	type
		PrefsObject = object(DObject)
				procedure Create (id: integer);
				override;
				procedure Destroy;
				override;
				procedure DoClose;
				override;
				procedure DoItem (item: integer);
				override;
			end;

	var
		noAlert: boolean;
		prefs_dialog_dirty: boolean;
		prefs_object: PrefsObject;

	procedure SetDialog (dp: dialogPtr);
	forward;

	procedure SetPrefsMenu (themenu, theitem: integer);
	begin
		SetIDItemEnable(themenu, theitem, not IsWObjectFront(prefs_object));
	end;

	procedure SetControls (dp: dialogPtr; first, count, value: integer);
		var
			i: integer;
	begin
		for i := 0 to count - 1 do
			SetDCtlBoolean(dp, first + i, i = value);
		prefs_dialog_dirty := true;
	end;

	function GetControls (dp: dialogPtr; first, count: integer): integer;
		var
			i: integer;
	begin
		GetControls := 0;
		for i := 0 to count - 1 do
			if GetDCtlBoolean(dp, first + i) then
				GetControls := i;
	end;

	procedure SetDialog (dp: dialogPtr);
	begin
		SetDCtlBoolean(dp, IPS_item, prefs.GetTagBoolean(P_showIP));
		SetDCtlBoolean(dp, AFS_item, prefs.GetTagBoolean(P_AutoOpen));
		SetDCtlBoolean(dp, Paid_item, prefs.GetTagBoolean(P_Paid));
		prefs_dialog_dirty := false;
	end;

	procedure GetDialog (dp: dialogPtr);
	begin
		prefs.SetTagBoolean(P_showIP, GetDCtlBoolean(dp, IPS_item));
		prefs.SetTagBoolean(P_AutoOpen, GetDCtlBoolean(dp, AFS_item));
		prefs.SetTagBoolean(P_Paid, GetDCtlBoolean(dp, Paid_item));
		UpdateAboutBox;
		prefs_dialog_dirty := false;
	end;

	procedure OpenPreferences;
	begin
		if prefs_object <> nil then begin
			if FrontWindow <> prefs_object.window then
				SelectWindow(prefs_object.window);
		end
		else begin
			new(prefs_object);
			prefs_object.Create(pref_dialog_id);
		end;
	end;

	procedure PrefsObject.Create (id: integer);
	begin
		UseResFile(app_resfile);
		SetVersionParamText('');
		inherited Create(id);
		SetOOOutline(Pref_Default_Item, User_item);
		cancel_item := Pref_Cancel_Item;
		SetDialog(window);
		ShowWindow(window);
	end;

	procedure Toggle (dp: dialogPtr; item: integer);
	begin
		SetDCtlBoolean(dp, item, not GetDCtlBoolean(dp, item));
		prefs_dialog_dirty := true;
	end;

	procedure PrefsObject.DoItem (item: integer);
		var
			reply: MySFReply;
			typeList: SFTypeList;
			dfs: boolean;
	begin
		case item of
			IPS_item, AFS_item, Paid_item: 
				Toggle(window, item);
			Pref_Save_item, Pref_Cancel_item:  begin
				if prefs_dialog_dirty and (item = Pref_Save_item) then begin
					GetDialog(window);
					WritePrefsData;
				end;
				Destroy;
			end;
			otherwise
				;
		end;
	end;

	procedure PrefsObject.Destroy;
	begin
		inherited Destroy;
		prefs_object := nil;
	end;

	procedure PrefsObject.DoClose;
		var
			dp: dialogPtr;
			sc: SCType;
	begin
		if prefs_dialog_dirty then begin
			if FrontWindow <> window then
				SelectWindow(window);
			sc := SaveChanges;
			if sc = SCSave then begin
				GetDialog(window);
				WritePrefsData;
			end;
		end
		else begin
			sc := SCDiscard;
		end;
		if sc <> SCCancel then begin
			Destroy;
		end;
	end;

	procedure InitPreferencesR;
	begin
		prefs_object := nil;
		noAlert := false;
		SetFBoth(Cpreferences, @OpenPreferences, @SetPrefsMenu);
	end;

	function FinishPreferencesR: boolean;
	begin
		if prefs_object <> nil then
			prefs_object.DoClose;
		FinishPreferencesR := prefs_object <> nil;
	end;

end.